from typing import Optional, Union

from supervisely.app import StateJson, DataJson
from supervisely.app.widgets import Widget


class Badge(Widget):
    def __init__(
        self,
        value: Union[int, str, float] = None,
        widget: Optional[Widget] = None,
        max: Union[int, float] = None,
        is_dot: bool = False,
        hidden: bool = False,
        widget_id: str = None,
    ):
        self._value = value
        self._widget = widget
        self._max = max if type(max) in [int, float] else None
        self._hidden = hidden
        self._is_dot = is_dot

        if self._value is None and self._widget is not None:
            self._is_dot = True

        if self._is_dot is True and self._value is None:
            self._value = 0

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        res = {}
        res["max"] = self._max
        res["isDot"] = self._is_dot
        res["hidden"] = self._hidden
        return res

    def get_json_state(self):
        return {"value": self._value}

    def set_value(self, value: Union[str, int, float]) -> None:
        self._value = value
        StateJson()[self.widget_id]["value"] = self._value
        StateJson().send_changes()

    def get_value(self) -> Union[str, int, float]:
        if "value" not in StateJson()[self.widget_id].keys():
            return None
        value = StateJson()[self.widget_id]["value"]
        return value

    def clear(self):
        self._value = None
        StateJson()[self.widget_id]["value"] = self._value
        StateJson().send_changes()

    def hide_badge(self):
        self._hidden = True
        DataJson()[self.widget_id]["hidden"] = self._hidden
        DataJson().send_changes()

    def show_badge(self):
        self._hidden = False
        DataJson()[self.widget_id]["hidden"] = self._hidden
        DataJson().send_changes()

    def toggle_visibility(self):
        self._hidden = not self._hidden
        DataJson()[self.widget_id]["hidden"] = self._hidden
        DataJson().send_changes()



# coding: utf-8

from copy import deepcopy
from re import L
from typing import List
import uuid
from supervisely.volume_annotation.volume_figure import VolumeFigure

from supervisely.project.project_meta import ProjectMeta
from supervisely._utils import take_with_default
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.volume_annotation.slice import Slice
from supervisely.volume_annotation.volume_tag_collection import VolumeTagCollection
from supervisely.volume_annotation.volume_object_collection import (
    VolumeObjectCollection,
)
from supervisely.volume_annotation.plane import Plane
from supervisely.volume_annotation.constants import (
    NAME,
    TAGS,
    OBJECTS,
    KEY,
    VOLUME_ID,
    VOLUME_META,
    PLANES,
    SPATIAL_FIGURES,
)


class VolumeAnnotation:
    """
    VolumeAnnotation for a single volume. :class:`VolumeAnnotation<VolumeAnnotation>` object is immutable.

    :param volume_meta: Metadata of the volume.
    :type volume_meta: dict
    :param objects: VolumeObjectCollection object.
    :type objects: VolumeObjectCollection, optional
    :param plane_sagittal: Sagittal plane of the volume.
    :type plane_sagittal: :py:class:`Plane<supervisely.volume_annotation.plane.Plane>`, optional
    :param plane_coronal: Coronal plane of the volume.
    :type plane_coronal: :py:class:`Plane<supervisely.volume_annotation.plane.Plane>`, optional
    :param plane_axial: Axial plane of the volume.
    :type plane_axial: :py:class:`Plane<supervisely.volume_annotation.plane.Plane>`, optional
    :param tags: VolumeTagCollection object.
    :type tags: VolumeTagCollection, optional
    :param spatial_figures: List of spatial figures associated with the volume.
    :type spatial_figures: List[VolumeFigure], optional
    :param key: UUID object.
    :type key: UUID, optional
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # Simple VolumeAnnotation example
        path = "/home/admin/work/volumes/vol_01.nrrd"
        volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
        volume_ann = sly.VolumeAnnotation(volume_meta)
        print(volume_ann.to_json())
        # Output: {
        # {
        #     "key": "56107223943346e5900fc256b8dcd7f0",
        #     "objects": [],
        #     "planes": [
        #         { "name": "sagittal", "normal": { "x": 1, "y": 0, "z": 0 }, "slices": [] },
        #         { "name": "coronal", "normal": { "x": 0, "y": 1, "z": 0 }, "slices": [] },
        #         { "name": "axial", "normal": { "x": 0, "y": 0, "z": 1 }, "slices": [] }
        #     ],
        #     "spatialFigures": [],
        #     "tags": [],
        #     "volumeMeta": {
        #         "ACS": "RAS",
        #         "channelsCount": 1,
        #         "dimensionsIJK": { "x": 512, "y": 512, "z": 139 },
        #         "directions": [1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0],
        #         "intensity": { "max": 3071.0, "min": -3024.0 },
        #         "origin": [-194.238403081894, -217.5384061336518, -347.7500000000001],
        #         "rescaleIntercept": 0,
        #         "rescaleSlope": 1,
        #         "spacing": [0.7617189884185793, 0.7617189884185793, 2.5],
        #         "windowCenter": 23.5,
        #         "windowWidth": 6095.0
        #     }
        # }

        # More complex VolumeAnnotation example

        path = "/home/admin/work/volumes/vol_01.nrrd"
        volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
        # VolumeObjectCollection
        obj_class = sly.ObjClass('brain', sly.Rectangle)
        volume_obj = sly.VolumeObject(obj_class)
        objects = sly.VolumeObjectCollection([volume_obj])
        # VolumeTagCollection
        brain_meta = sly.TagMeta('brain_tag', sly.TagValueType.ANY_STRING)
        from supervisely.volume_annotation.volume_tag import VolumeTag
        vol_tag = VolumeTag(brain_meta, value='human')
        from supervisely.volume_annotation.volume_tag_collection import VolumeTagCollection
        volume_tags = VolumeTagCollection([vol_tag])

        volume_ann = sly.VolumeAnnotation(volume_meta, objects, volume_tags)
        print(volume_ann.to_json())
        # Output:
        # {
        #     "key": "4d4bb69e6fcd40e1a1cb076c07769903",
        #     "objects": [
        #         {
        #         "classTitle": "brain",
        #         "key": "22e1082a17f74279b00eed0bfb0ba11d",
        #         "tags": []
        #         }
        #     ],
        #     "planes": [
        #         { "name": "sagittal", "normal": { "x": 1, "y": 0, "z": 0 }, "slices": [] },
        #         { "name": "coronal", "normal": { "x": 0, "y": 1, "z": 0 }, "slices": [] },
        #         { "name": "axial", "normal": { "x": 0, "y": 0, "z": 1 }, "slices": [] }
        #     ],
        #     "spatialFigures": [],
        #     "tags": [
        #         {
        #         "key": "b9de6631d328441796119b4b0039fc61",
        #         "name": "brain_tag",
        #         "value": "human"
        #         }
        #     ],
        #     "volumeMeta": {
        #         "ACS": "RAS",
        #         "channelsCount": 1,
        #         "dimensionsIJK": { "x": 512, "y": 512, "z": 139 },
        #         "directions": [1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0],
        #         "intensity": { "max": 3071.0, "min": -3024.0 },
        #         "origin": [-194.238403081894, -217.5384061336518, -347.7500000000001],
        #         "rescaleIntercept": 0,
        #         "rescaleSlope": 1,
        #         "spacing": [0.7617189884185793, 0.7617189884185793, 2.5],
        #         "windowCenter": 23.5,
        #         "windowWidth": 6095.0
        #     }
        # }
    """

    def __init__(
        self,
        volume_meta,
        objects=None,
        plane_sagittal=None,
        plane_coronal=None,
        plane_axial=None,
        tags=None,
        spatial_figures=None,
        key=None,
    ):
        self._volume_meta = volume_meta
        self._tags = take_with_default(tags, VolumeTagCollection())
        self._objects = take_with_default(objects, VolumeObjectCollection())
        self._key = take_with_default(key, uuid.uuid4())

        self._plane_sagittal = take_with_default(
            plane_sagittal,
            Plane(Plane.SAGITTAL, volume_meta=volume_meta),
        )
        self._plane_coronal = take_with_default(
            plane_coronal,
            Plane(Plane.CORONAL, volume_meta=volume_meta),
        )
        self._plane_axial = take_with_default(
            plane_axial,
            Plane(Plane.AXIAL, volume_meta=volume_meta),
        )

        self._spatial_figures = take_with_default(spatial_figures, [])
        self.validate_figures_bounds()

    @property
    def volume_meta(self) -> dict:
        """
        Volume meta data.

        :returns: Sagittal plane of the volume.
        :rtype: dict

        :Usage example:

         .. code-block:: python

            import supervisely as sly
            path = "/home/admin/work/volumes/vol_01.nrrd"
            volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
            vol_ann = sly.VolumeAnnotation(volume_meta)
            volume_meta = vol_ann.volume_meta
        """

        return deepcopy(self._volume_meta)

    @property
    def plane_sagittal(self) -> Plane:
        """
        Sagital plane of the volume.

        :returns: Sagittal plane of the volume.
        :rtype: :py:class:`Plane<supervisely.volume_annotation.plane.Plane>`

        :Usage example:

         .. code-block:: python

            import supervisely as sly
            path = "/home/admin/work/volumes/vol_01.nrrd"
            volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
            vol_ann = sly.VolumeAnnotation(volume_meta)
            plane_sagittal = vol_ann.plane_sagittal
        """

        return self._plane_sagittal

    @property
    def plane_coronal(self) -> Plane:
        """
        Coronal plane of the volume.

        :returns: Coronal plane of the volume.
        :rtype: :py:class:`Plane<supervisely.volume_annotation.plane.Plane>`

        :Usage example:

         .. code-block:: python

            import supervisely as sly
            path = "/home/admin/work/volumes/vol_01.nrrd"
            volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
            vol_ann = sly.VolumeAnnotation(volume_meta)
            plane_coronal = vol_ann.plane_coronal
        """

        return self._plane_coronal

    @property
    def plane_axial(self) -> Plane:
        """
        Axial plane of the volume.

        :returns: Axial plane of the volume.
        :rtype: :py:class:`Plane<supervisely.volume_annotation.plane.Plane>`

        :Usage example:

         .. code-block:: python

            import supervisely as sly
            path = "/home/admin/work/volumes/vol_01.nrrd"
            volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
            vol_ann = sly.VolumeAnnotation(volume_meta)
            plane_axial = vol_ann.plane_axial
        """

        return self._plane_axial

    @property
    def objects(self) -> VolumeObjectCollection:
        """
        VolumeAnnotation objects.

        :return: VolumeObjectCollection object
        :rtype: :class:`VolumeObjectCollection<supervisely.volume_annotation.volume_object_collection.VolumeObjectCollection>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            path = "/Users/Downloads/volumes/Demo volumes_ds1_CTChest.nrrd"
            volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)

            # VolumeObjectCollection
            obj_class_heart = sly.ObjClass('heart', sly.Rectangle)
            volume_obj_heart = sly.VolumeObject(obj_class_heart)
            objects = sly.VolumeObjectCollection([volume_obj_heart])
            volume_ann = sly.VolumeAnnotation(volume_meta, objects)

            print(volume_ann.objects.to_json())
            # Output: [
            #     {
            #         "key": "2b5d70baa5a74d06a525b950b5f2b756",
            #         "classTitle": "heart",
            #         "tags": []
            #     }
            # ]
        """

        return self._objects

    @property
    def tags(self) -> VolumeTagCollection:
        """
        VolumeTag objects.

        :returns: VolumeTagCollection
        :rtype: :class:`VolumeTagCollection<supervisely.volume_annotation.volume_tag_collection.VolumeTagCollection>`

        :Usage example:

         .. code-block:: python

            import supervisely as sly
            path = "/home/admin/work/volumes/vol_01.nrrd"
            volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
            vol_ann = sly.VolumeAnnotation(volume_meta)
            tags = vol_ann.tags
        """

        return self._tags

    @property
    def spatial_figures(self) -> List[VolumeFigure]:
        """
        Get a list of spatial figures.

        :returns: List of spatial figures from VolumeAnnotation object.
        :rtype: List[VolumeFigure]

        :Usage example:

         .. code-block:: python

            import supervisely as sly
            path = "/home/admin/work/volumes/vol_01.nrrd"
            volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
            vol_ann = sly.VolumeAnnotation(volume_meta)
            spatial_figures = vol_ann.spatial_figures
        """

        return self._spatial_figures

    @property
    def figures(self) -> List[VolumeFigure]:
        """
        VolumeFigure objects.

        :returns: List of VolumeFigure objects from VolumeAnnotation object.
        :rtype: List[VolumeFigure]

        :Usage example:

         .. code-block:: python

            import supervisely as sly
            path = "/home/admin/work/volumes/vol_01.nrrd"
            volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
            vol_ann = sly.VolumeAnnotation(volume_meta)
            figures = vol_ann.figures
        """

        all_figures = []
        for plane in [self.plane_sagittal, self.plane_coronal, self.plane_axial]:
            all_figures.extend(plane.figures)
        return all_figures

    def key(self) -> str:
        """
        Volume annotation key value.

        :returns: Key value of VolumeAnnotation object.
        :rtype: str

        :Usage example:

         .. code-block:: python

            import supervisely as sly
            path = "/home/admin/work/volumes/vol_01.nrrd"
            volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
            vol_ann = sly.VolumeAnnotation(volume_meta)
            key = vol_ann.key()
        """
        return self._key

    def validate_figures_bounds(self):
        """
        Checks if all slices in each plane contains figures.

        :raises: :class:`OutOfImageBoundsException<supervisely.video_annotation.video_figure.OutOfImageBoundsException>`, if figure is out of slices images bounds
        :return: None
        :rtype: :class:`NoneType`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            plane_axial = sly.Plane(sly.Plane.AXIAL, frames, volume_meta=volume_meta)
            volume_ann = sly.VolumeAnnotation(volume_meta, objects, plane_axial=plane_axial)
            volume_ann.validate_figures_bounds()
        """

        self.plane_sagittal.validate_figures_bounds()
        self.plane_coronal.validate_figures_bounds()
        self.plane_axial.validate_figures_bounds()

    def is_empty(self) -> bool:
        """
        Check whether volume annotation contains objects or tags, or not.

        :returns: True if volume annotation is empty, False otherwise.
        :rtype: :class:`bool`

        :Usage exmaple:

         .. code-block:: python

            import supervisely as sly
            path = "/home/admin/work/volumes/vol_01.nrrd"
            volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
            vol_ann = sly.VolumeAnnotation(volume_meta)

            is_empty = vol_ann.is_empty()
        """

        if len(self.objects) == 0 and len(self.tags) == 0:
            return True
        else:
            return False

    def clone(
        self,
        volume_meta=None,
        objects=None,
        plane_sagittal=None,
        plane_coronal=None,
        plane_axial=None,
        tags=None,
        spatial_figures=None,
    ):
        """
        Makes a copy of VolumeAnnotation with new fields, if fields are given, otherwise it will use fields of the original VolumeAnnotation.

        :param volume_meta: Metadata of the volume.
        :type volume_meta: dict
        :param objects: VolumeObjectCollection object.
        :type objects: VolumeObjectCollection, optional
        :param plane_sagittal: Sagittal plane of the volume.
        :type plane_sagittal: :py:class:`Plane<supervisely.volume_annotation.plane.Plane>`, optional
        :param plane_coronal: Coronal plane of the volume.
        :type plane_coronal: :py:class:`Plane<supervisely.volume_annotation.plane.Plane>`, optional
        :param plane_axial: Axial plane of the volume.
        :type plane_axial: :py:class:`Plane<supervisely.volume_annotation.plane.Plane>`, optional
        :param tags: VolumeTagCollection object.
        :type tags: VolumeTagCollection, optional
        :param spatial_figures: List of spatial figures associated with the volume.
        :type spatial_figures: List[VolumeFigure], optional

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            path = "/home/admin/work/volumes/vol_01.nrrd"
            volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
            volume_ann = sly.VolumeAnnotation(volume_meta)

            obj_class_heart = sly.ObjClass('heart', sly.Rectangle)
            video_obj_heart = sly.VolumeObject(obj_class_heart)
            new_objects = sly.VolumeObjectCollection([volume_obj_heart])
            new_volume_ann = volume_ann.clone(objects=new_objects)
        """

        return VolumeAnnotation(
            volume_meta=take_with_default(volume_meta, self.volume_meta),
            objects=take_with_default(objects, self.objects),
            plane_sagittal=take_with_default(plane_sagittal, self.plane_sagittal),
            plane_coronal=take_with_default(plane_coronal, self.plane_coronal),
            plane_axial=take_with_default(plane_axial, self.plane_axial),
            tags=take_with_default(tags, self.tags),
            spatial_figures=take_with_default(spatial_figures, self.spatial_figures),
        )

    @classmethod
    def from_json(cls, data: dict, project_meta: ProjectMeta, key_id_map: KeyIdMap = None):
        """
        Convert a json dict to VolumeAnnotation.

        :param data: Volume annotation in json format as a dict.
        :type data: dict
        :param project_meta: Input :class:`ProjectMeta<supervisely.project.project_meta.ProjectMeta>`.
        :type project_meta: ProjectMeta
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :return: VolumeAnnotation object
        :rtype: :class:`VolumeAnnotation<VolumeAnnotation>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            from supervisely.video_annotation.key_id_map import KeyIdMap

            meta = sly.ProjectMeta()
            key_id_map = KeyIdMap()

            ann_json = {
                "key": "56107223943346e5900fc256b8dcd7f0",
                "objects": [],
                "planes": [
                    { "name": "sagittal", "normal": { "x": 1, "y": 0, "z": 0 }, "slices": [] },
                    { "name": "coronal", "normal": { "x": 0, "y": 1, "z": 0 }, "slices": [] },
                    { "name": "axial", "normal": { "x": 0, "y": 0, "z": 1 }, "slices": [] }
                ],
                "spatialFigures": [],
                "tags": [],
                "volumeMeta": {
                    "ACS": "RAS",
                    "channelsCount": 1,
                    "dimensionsIJK": { "x": 512, "y": 512, "z": 139 },
                    "directions": [1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0],
                    "intensity": { "max": 3071.0, "min": -3024.0 },
                    "origin": [-194.238403081894, -217.5384061336518, -347.7500000000001],
                    "rescaleIntercept": 0,
                    "rescaleSlope": 1,
                    "spacing": [0.7617189884185793, 0.7617189884185793, 2.5],
                    "windowCenter": 23.5,
                    "windowWidth": 6095.0
                }
            }


            ann = sly.VolumeAnnotation.from_json(ann_json, meta, key_id_map)
        """

        volume_key = uuid.UUID(data[KEY]) if KEY in data else uuid.uuid4()
        if key_id_map is not None:
            key_id_map.add_video(volume_key, data.get(VOLUME_ID, None))

        volume_meta = data[VOLUME_META]

        tags = VolumeTagCollection.from_json(data[TAGS], project_meta.tag_metas, key_id_map)
        objects = VolumeObjectCollection.from_json(data[OBJECTS], project_meta, key_id_map)

        plane_sagittal = None
        plane_coronal = None
        plane_axial = None
        for plane_json in data[PLANES]:
            if plane_json[NAME] == Plane.SAGITTAL:
                plane_sagittal = Plane.from_json(
                    plane_json,
                    Plane.SAGITTAL,
                    objects,
                    volume_meta=volume_meta,
                    key_id_map=key_id_map,
                )
            elif plane_json[NAME] == Plane.CORONAL:
                plane_coronal = Plane.from_json(
                    plane_json,
                    Plane.CORONAL,
                    objects,
                    volume_meta=volume_meta,
                    key_id_map=key_id_map,
                )
            elif plane_json[NAME] == Plane.AXIAL:
                plane_axial = Plane.from_json(
                    plane_json,
                    Plane.AXIAL,
                    objects,
                    volume_meta=volume_meta,
                    key_id_map=key_id_map,
                )
            else:
                raise RuntimeError(f"Unknown plane name {plane_json[NAME]}")

        spatial_figures = []
        for figure_json in data.get(SPATIAL_FIGURES, []):
            figure = VolumeFigure.from_json(
                figure_json,
                objects,
                plane_name=None,
                slice_index=None,
                key_id_map=key_id_map,
            )
            spatial_figures.append(figure)

        return cls(
            volume_meta=volume_meta,
            objects=objects,
            plane_sagittal=plane_sagittal,
            plane_coronal=plane_coronal,
            plane_axial=plane_axial,
            tags=tags,
            spatial_figures=spatial_figures,
            key=volume_key,
        )

    def to_json(self, key_id_map: KeyIdMap = None) -> dict:
        """
        Convert the VolumeAnnotation to a json dict.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :return: Volume annotation in json format as a dict.
        :rtype: dict
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            from supervisely.video_annotation.key_id_map import KeyIdMap

            path = "/home/admin/work/volumes/vol_01.nrrd"
            volume, volume_meta = sly.volume.read_nrrd_serie_volume_np(path)
            volume_ann = sly.VolumeAnnotation(volume_meta)

            print(volume_ann.to_json())
            # Output: {
            # {
            #     "key": "56107223943346e5900fc256b8dcd7f0",
            #     "objects": [],
            #     "planes": [
            #         { "name": "sagittal", "normal": { "x": 1, "y": 0, "z": 0 }, "slices": [] },
            #         { "name": "coronal", "normal": { "x": 0, "y": 1, "z": 0 }, "slices": [] },
            #         { "name": "axial", "normal": { "x": 0, "y": 0, "z": 1 }, "slices": [] }
            #     ],
            #     "spatialFigures": [],
            #     "tags": [],
            #     "volumeMeta": {
            #         "ACS": "RAS",
            #         "channelsCount": 1,
            #         "dimensionsIJK": { "x": 512, "y": 512, "z": 139 },
            #         "directions": [1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0],
            #         "intensity": { "max": 3071.0, "min": -3024.0 },
            #         "origin": [-194.238403081894, -217.5384061336518, -347.7500000000001],
            #         "rescaleIntercept": 0,
            #         "rescaleSlope": 1,
            #         "spacing": [0.7617189884185793, 0.7617189884185793, 2.5],
            #         "windowCenter": 23.5,
            #         "windowWidth": 6095.0
            #     }
            # }
        """

        res_json = {
            VOLUME_META: self.volume_meta,
            KEY: self.key().hex,
            TAGS: self.tags.to_json(key_id_map),
            OBJECTS: self.objects.to_json(key_id_map),
            PLANES: [
                self.plane_sagittal.to_json(),
                self.plane_coronal.to_json(),
                self.plane_axial.to_json(),
            ],
            SPATIAL_FIGURES: [figure.to_json(key_id_map) for figure in self.spatial_figures],
        }

        if key_id_map is not None:
            volume_id = key_id_map.get_video_id(self.key())
            if volume_id is not None:
                res_json[VOLUME_ID] = volume_id

        return res_json



from typing import List, Optional, Callable, Union
from tqdm import tqdm

from supervisely.api.api import Api
from supervisely.project.project import download_project
from supervisely.project.pointcloud_episode_project import download_pointcloud_episode_project
from supervisely.project.pointcloud_project import download_pointcloud_project
from supervisely.project.video_project import download_video_project
from supervisely.project.volume_project import download_volume_project
from supervisely.project.project_type import ProjectType


def download(
    api: Api,
    project_id: int,
    dest_dir: str,
    dataset_ids: Optional[List[int]] = None,
    log_progress: Optional[bool] = False,
    progress_cb: Optional[Union[tqdm, Callable]] = None,
    **kwargs,
) -> None:
    """
    Downloads project of any type to the local directory. See methods `sly.download_project`,
    `sly.download_video_project`, `sly.download_volume_project`, `sly.download_pointcloud_project`,
    `sly.download_pointcloud_episode_project` to examine full list of possible arguments.

    :param api: Supervisely API address and token.
    :type api: Api
    :param project_id: Project ID, which will be downloaded.
    :type project_id: int
    :param dest_dir: Destination path to local directory.
    :type dest_dir: str
    :param dataset_ids: Specified list of Dataset IDs which will be downloaded. Datasets could be downloaded from different projects but with the same data type.
    :type dataset_ids: list(int), optional
    :param log_progress: Show downloading logs in the output.
    :type log_progress: bool, optional
    :param progress_cb: Function for tracking download progress.
    :type progress_cb: tqdm or callable, optional

    :return: None.
    :rtype: NoneType
    :Usage example:

    .. code-block:: python

        import os
        from dotenv import load_dotenv

        from tqdm import tqdm
        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervise.ly", token="4r47N...xaTatb")

        dest_dir = 'your/local/dest/dir'

        # Download image project
        project_id_image = 17732
        project_info = api.project.get_info_by_id(project_id_image)
        num_images = project_info.items_count

        p = tqdm(desc="Downloading image project", total=num_images)
        sly.download(
            api,
            project_id_image,
            dest_dir,
            progress_cb=p,
            save_image_info=True,
            save_images=True,
        )

        # Download video project
        project_id_video = 60498
        project_info = api.project.get_info_by_id(project_id_video)
        num_videos = project_info.items_count

        p = tqdm(desc="Downloading video project", total=num_videos)
        sly.download(
            api,
            project_id_video,
            dest_dir,
            progress_cb=p,
            save_video_info=True,
        )

        # Download volume project
        project_id_volume = 18594
        project_info = api.project.get_info_by_id(project_id_volume)
        num_volumes = project_info.items_count

        p = tqdm(desc="Downloading volume project",total=num_volumes)
        sly.download(
            api,
            project_id_volume,
            dest_dir,
            progress_cb=p,
            download_volumes=True,
        )

        # Download pointcloud project
        project_id_ptcl = 18592
        project_info = api.project.get_info_by_id(project_id_ptcl)
        num_ptcl = project_info.items_count

        p = tqdm(desc="Downloading pointcloud project", total=num_ptcl)
        sly.download(
            api,
            project_id_ptcl,
            dest_dir,
            progress_cb=p,
            download_pointclouds_info=True,
        )

        # Download some datasets from pointcloud episodes project
        project_id_ptcl_ep = 18593
        dataset_ids = [43546, 45765, 45656]

        p = tqdm(
            desc="Download some datasets from pointcloud episodes project",
            total=len(dataset_ids),
        )
        sly.download(
            api,
            project_id_ptcl_ep,
            dest_dir,
            dataset_ids,
            progress_cb=p,
            download_pcd=True,
            download_related_images=True,
            download_annotations=True,
            download_pointclouds_info=True,
        )
    """

    project_info = api.project.get_info_by_id(project_id)

    if project_info.type == ProjectType.IMAGES.value:
        download_project(
            api=api,
            project_id=project_id,
            dest_dir=dest_dir,
            dataset_ids=dataset_ids,
            log_progress=log_progress,
            progress_cb=progress_cb,
            **kwargs,
        )
    elif project_info.type == ProjectType.VIDEOS.value:
        download_video_project(
            api=api,
            project_id=project_id,
            dest_dir=dest_dir,
            dataset_ids=dataset_ids,
            log_progress=log_progress,
            progress_cb=progress_cb,
            **kwargs,
        )
    elif project_info.type == ProjectType.VOLUMES.value:
        download_volume_project(
            api=api,
            project_id=project_id,
            dest_dir=dest_dir,
            dataset_ids=dataset_ids,
            log_progress=log_progress,
            progress_cb=progress_cb,
            **kwargs,
        )

    elif project_info.type == ProjectType.POINT_CLOUDS.value:
        download_pointcloud_project(
            api=api,
            project_id=project_id,
            dest_dir=dest_dir,
            dataset_ids=dataset_ids,
            log_progress=log_progress,
            progress_cb=progress_cb,
            **kwargs,
        )
    elif project_info.type == ProjectType.POINT_CLOUD_EPISODES.value:
        download_pointcloud_episode_project(
            api=api,
            project_id=project_id,
            dest_dir=dest_dir,
            dataset_ids=dataset_ids,
            log_progress=log_progress,
            progress_cb=progress_cb,
            **kwargs,
        )
    else:
        raise ValueError(f"Unknown type of project ({project_info.type})")



from typing import List
from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget
from supervisely.sly_logger import logger


class Card(Widget):
    def __init__(
        self,
        title: str = None,
        description: str = None,
        collapsable: bool = False,
        content: Widget = None,
        content_top_right: Widget = None,
        lock_message="Card content is locked",
        widget_id: str = None,
    ):
        self._title = title
        self._description = description
        self._collapsable = collapsable
        self._collapsed = False
        self._content = content
        self._show_slot = False
        self._slot_content = content_top_right
        if self._slot_content is not None:
            self._show_slot = True
        self._options = {"collapsable": self._collapsable, "marginBottom": "0px"}
        self._lock_message = lock_message
        super().__init__(widget_id=widget_id, file_path=__file__)
        self._disabled = {"disabled": False, "message": self._lock_message}

    def get_json_data(self):
        return {
            "title": self._title,
            "description": self._description,
            "collapsable": self._collapsable,
            "options": self._options,
            "show_slot": self._show_slot,
        }

    def get_json_state(self):
        return {"disabled": self._disabled, "collapsed": self._collapsed}

    def collapse(self):
        if self._collapsable is False:
            logger.warn(f"Card {self.widget_id} can not be collapsed")
            return
        self._collapsed = True
        StateJson()[self.widget_id]["collapsed"] = self._collapsed
        StateJson().send_changes()

    def uncollapse(self):
        if self._collapsable is False:
            logger.warn(f"Card {self.widget_id} can not be uncollapsed")
            return
        self._collapsed = False
        StateJson()[self.widget_id]["collapsed"] = self._collapsed
        StateJson().send_changes()

    def lock(self, message: str = None):
        if message is not None:
            self._lock_message = message
        self._disabled = {"disabled": True, "message": self._lock_message}
        StateJson()[self.widget_id]["disabled"] = self._disabled
        StateJson().send_changes()

    def unlock(self):
        self._disabled["disabled"] = False
        StateJson()[self.widget_id]["disabled"] = self._disabled
        StateJson().send_changes()

    def is_locked(self):
        return self._disabled["disabled"]



# coding: utf-8

import numpy as np

from supervisely.metric.metric_base import MetricsBase
from supervisely.metric.common import render_labels_for_classes, safe_ratio, TRUE_POSITIVE, TRUE_NEGATIVE, \
    FALSE_POSITIVE, FALSE_NEGATIVE, PRECISION, RECALL, ACCURACY, TOTAL


class PixelAccuracyMetric(MetricsBase):
    def __init__(self, class_mapping):
        self._class_mapping = class_mapping.copy()
        if len(self._class_mapping) < 1:
            raise RuntimeError('At least one classes pair should be defined.')
        self._counters = {cls_gt: {TRUE_POSITIVE: 0, TRUE_NEGATIVE: 0, FALSE_POSITIVE: 0, FALSE_NEGATIVE: 0, TOTAL: 0}
                          for cls_gt in self._class_mapping.keys()}
        self._total_counters = {TRUE_POSITIVE: 0, TRUE_NEGATIVE: 0, FALSE_POSITIVE: 0, FALSE_NEGATIVE: 0, TOTAL: 0}
        self._gt_colors = {}
        self._pred_colors = {}
        for class_idx, (gt_class, pred_class) in enumerate(self._class_mapping.items(), start=1):
            self._gt_colors[gt_class] = class_idx
            self._pred_colors[pred_class] = class_idx

    def add_pair(self, ann_gt, ann_pred):
        img_size = ann_gt.img_size
        render_gt, render_pred = np.zeros(img_size, dtype=np.int32), np.zeros(img_size, dtype=np.int32)
        render_labels_for_classes(ann_gt.labels, self._gt_colors, render_gt, missing_classes_color=0)
        render_labels_for_classes(ann_pred.labels, self._pred_colors, render_pred, missing_classes_color=0)

        for cls_gt, cls_pred in self._class_mapping.items():
            mask_gt = (render_gt == self._gt_colors[cls_gt])
            mask_pred = (render_pred == self._pred_colors[cls_pred])
            class_pair_counters = self._counters[cls_gt]
            class_pair_counters[TRUE_POSITIVE] += (mask_gt & mask_pred).sum()
            class_pair_counters[TRUE_NEGATIVE] += (np.logical_not(mask_gt) & np.logical_not(mask_pred)).sum()
            class_pair_counters[FALSE_POSITIVE] += (np.logical_not(mask_gt) & mask_pred).sum()
            class_pair_counters[FALSE_NEGATIVE] += (mask_gt & np.logical_not(mask_pred)).sum()
            class_pair_counters[TOTAL] += mask_gt.size

        self._total_counters[TRUE_POSITIVE] += ((render_gt == render_pred) & (render_pred != 0)).sum()
        self._total_counters[TRUE_NEGATIVE] += ((render_gt == 0) & (render_pred == 0)).sum()
        self._total_counters[FALSE_POSITIVE] += ((render_gt != render_pred) & (render_pred != 0)).sum()
        self._total_counters[FALSE_NEGATIVE] += ((render_gt != 0) & (render_pred == 0)).sum()
        self._total_counters[TOTAL] += render_gt.size

    @staticmethod
    def _metrics_from_counters(counters):
        return {PRECISION: safe_ratio(counters[TRUE_POSITIVE], counters[TRUE_POSITIVE] + counters[FALSE_POSITIVE]),
                RECALL: safe_ratio(counters[TRUE_POSITIVE], counters[TRUE_POSITIVE] + counters[FALSE_NEGATIVE]),
                ACCURACY: safe_ratio(counters[TRUE_POSITIVE] + counters[TRUE_NEGATIVE], counters[TOTAL])}

    def get_metrics(self):
        return {cls_gt: self._metrics_from_counters(class_counters) for cls_gt, class_counters in
                self._counters.items()}

    def get_total_metrics(self):
        return self._metrics_from_counters(self._total_counters)



import traceback

from rich.console import Console
from tqdm import tqdm
import supervisely as sly


def download_run(id: int, dest_dir: str) -> bool:

    api = sly.Api.from_env()
    console = Console()

    console.print(
        f"\nDownloading data from project with ID={id} to directory: '{dest_dir}' ...\n",
        style="bold",
    )

    project_info = api.project.get_info_by_id(id)

    if project_info is None:
        console.print("\nError: Project not exists\n", style="bold red")
        return False

    n_count = project_info.items_count
    try:
        if sly.is_development():
            with tqdm(total=n_count) as pbar:
                sly.download(api, id, dest_dir, progress_cb=pbar.update)
                pbar.refresh()
        else:
            sly.download(api, id, dest_dir, log_progress=True)

        console.print("\nProject is downloaded sucessfully!\n", style="bold green")
        return True
    except:
        console.print(f"\nProject is not downloaded\n", style="bold red")
        traceback.print_exc()
        return False



import json
import os
import mimetypes
from os.path import basename, join

import supervisely.io.env as env
from supervisely._utils import is_production
from supervisely.api.api import Api
from supervisely.app.fastapi import get_name_from_env
from supervisely.io.fs import get_file_name_with_ext, silent_remove, archive_directory, remove_dir
import supervisely.io.env as sly_env
from supervisely import rand_str
from supervisely.task.progress import Progress
from supervisely.sly_logger import logger
from supervisely.team_files import RECOMMENDED_EXPORT_PATH


def set_project(id: int):
    if is_production() is True:
        api = Api()
        task_id = sly_env.task_id()
        api.task.set_output_project(task_id, project_id=id)
    else:
        print(f"Output project: id={id}")


def set_directory(teamfiles_dir: str):
    """
    Sets a link to a teamfiles directory in workspace tasks interface
    """
    if is_production():
        api = Api()
        task_id = sly_env.task_id()

        if api.task.get_info_by_id(task_id) is None:
            raise KeyError(
                f"Task with ID={task_id} is either not exist or not found in your account"
            )

        team_id = api.task.get_info_by_id(task_id)["teamId"]

        if api.team.get_info_by_id(team_id) is None:
            raise KeyError(
                f"Team with ID={team_id} is either not exist or not found in your account"
            )

        files = api.file.list2(team_id, teamfiles_dir, recursive=True)

        # if directory is empty or not exists
        if len(files) == 0:
            # some data to create dummy .json file to get file id
            data = {"team_id": team_id, "task_id": task_id, "directory": teamfiles_dir}
            filename = f"{rand_str(10)}.json"

            src_path = os.path.join("/tmp/", filename)
            with open(src_path, "w") as f:
                json.dump(data, f)

            dst_path = os.path.join(teamfiles_dir, filename)
            file_id = api.file.upload(team_id, src_path, dst_path).id

            silent_remove(src_path)

        else:
            file_id = files[0].id

        api.task.set_output_directory(task_id, file_id, teamfiles_dir)

    else:
        print(f"Output directory: '{teamfiles_dir}'")


def set_download(local_path: str):
    """
    Receives a path to the local file or directory. If the path is a directory, it will be archived before uploading.
    After sets a link to a uploaded file in workspace tasks interface according to the file type.
    If the file is an archive, the set_output_archive method is called and "Download archive" text is displayed.
    If the file is not an archive, the set_output_file_download method is called and "Download file" text is displayed.

    :param local_path: path to the local file or directory, which will be uploaded to the teamfiles
    :type local_path: str
    :return: None
    :rtype: None
    """
    if os.path.isdir(local_path):
        archive_path = f"{local_path}.tar"
        archive_directory(local_path, archive_path)
        remove_dir(local_path)
        local_path = archive_path

    if is_production():
        api = Api()
        task_id = sly_env.task_id()
        upload_progress = []

        team_id = env.team_id()

        def _print_progress(monitor, upload_progress):
            if len(upload_progress) == 0:
                upload_progress.append(
                    Progress(
                        message=f"Uploading '{basename(local_path)}'",
                        total_cnt=monitor.len,
                        ext_logger=logger,
                        is_size=True,
                    )
                )
            upload_progress[0].set_current_value(monitor.bytes_read)

        def _is_archive(local_path: str) -> bool:
            """
            Checks if the file is an archive by its mimetype using list of the most common archive mimetypes.

            :param local_path: path to the local file
            :type local_path: str
            :return: True if the file is an archive, False otherwise
            :rtype: bool
            """
            archive_mimetypes = [
                "application/zip",
                "application/x-tar",
                "application/x-gzip",
                "application/x-bzip2",
                "application/x-7z-compressed",
                "application/x-rar-compressed",
                "application/x-xz",
                "application/x-lzip",
                "application/x-lzma",
                "application/x-lzop",
                "application/x-bzip",
                "application/x-bzip2",
                "application/x-compress",
                "application/x-compressed",
            ]

            return mimetypes.guess_type(local_path)[0] in archive_mimetypes

        remote_path = join(
            RECOMMENDED_EXPORT_PATH,
            get_name_from_env(),
            str(task_id),
            f"{get_file_name_with_ext(local_path)}",
        )
        file_info = api.file.upload(
            team_id=team_id,
            src=local_path,
            dst=remote_path,
            progress_cb=lambda m: _print_progress(m, upload_progress),
        )

        if _is_archive(local_path):
            api.task.set_output_archive(task_id, file_info.id, file_info.name)
        else:
            api.task.set_output_file_download(task_id, file_info.id, file_info.name)

        logger.info(f"Remote file: id={file_info.id}, name={file_info.name}")
        silent_remove(local_path)

    else:
        print(f"Output file: '{local_path}'")



# coding: utf-8
"""base class for :class:`ObjClassCollection<supervisely.annotation.obj_class_collection.ObjClassCollection>`, :class:`TagMetaCollection<supervisely.annotation.tag_meta_collection.TagMetaCollection>`  and :class:`TagCollection<supervisely.annotation.tag_collection.TagCollection>` instances"""


# docs
from __future__ import annotations
from typing import List, Dict, Optional, Any


from prettytable import PrettyTable
from supervisely._utils import take_with_default
from typing import List, Iterable
from collections import defaultdict


class DuplicateKeyError(KeyError):
    r"""Raised when trying to add already existing key to
    :class:`KeyIndexedCollection <supervisely.collection.key_indexed_collection.KeyIndexedCollection>`"""
    pass


class KeyObject:
    """
    Base class fo objects that should implement ``key`` method. Child classes then can be stored in KeyIndexedCollection.
    """
    def key(self):
        raise NotImplementedError()


class KeyIndexedCollection:
    """
    Base class for :class:`ObjClassCollection<supervisely.annotation.obj_class_collection.ObjClassCollection>`, :class:`TagMetaCollection<supervisely.annotation.tag_meta_collection.TagMetaCollection>`  and :class:`TagCollection<supervisely.annotation.tag_collection.TagCollection>` instances.
    It is an analogue of python's standard Dict. It allows to store objects inherited from :class:`KeyObject <supervisely.collection.key_indexed_collection.KeyObject>`.

    :param items: List of :class:`ObjClassCollection<supervisely.annotation.obj_class_collection.ObjClassCollection>`, :class:`TagMetaCollection<supervisely.annotation.tag_meta_collection.TagMetaCollection>`  and :class:`TagCollection<supervisely.annotation.tag_collection.TagCollection>` objects.
    :type items: list, optional
    :raises: :class:`DuplicateKeyError<supervisely.collection.key_indexed_collection.DuplicateKeyError>`, when trying to add object with already existing key

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        item_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
        item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
        collection = sly.collection.key_indexed_collection.KeyIndexedCollection([item_cat, item_turtle])
        print(collection.to_json())
        # Output: [
        #     {
        #         "name": "cat",
        #         "value_type": "none",
        #         "color": "#8A0F12",
        #         "hotkey": "",
        #         "applicable_type": "all",
        #         "classes": []
        #     },
        #     {
        #         "name": "turtle",
        #         "value_type": "any_string",
        #         "color": "#8A860F",
        #         "hotkey": "",
        #         "applicable_type": "all",
        #         "classes": []
        #     }
        # ]

        # Try to add item with a key that already exists in the collection
        dublicate_item = sly.ObjClass('cat', sly.Rectangle)
        new_collection = collection.add(dublicate_item)
        # Output:
        # supervisely.collection.key_indexed_collection.DuplicateKeyError: "Key 'cat' already exists"

        # Add item with a key that not exist in the collection
        item_dog = sly.ObjClass('dog', sly.Rectangle)
        new_collection = collection.add(item_dog)
        print(new_collection.to_json())
        # Output: [
        #     {
        #         "name": "cat",
        #         "value_type": "none",
        #         "color": "#668A0F",
        #         "hotkey": "",
        #         "applicable_type": "all",
        #         "classes": []
        #     },
        #     {
        #         "name": "turtle",
        #         "value_type": "any_string",
        #         "color": "#4D0F8A",
        #         "hotkey": "",
        #         "applicable_type": "all",
        #         "classes": []
        #     },
        #     {
        #         "title": "dog",
        #         "shape": "rectangle",
        #         "color": "#0F7F8A",
        #         "geometry_config": {},
        #         "hotkey": ""
        #     }
        # ]
    """

    item_type = KeyObject
    """
    The type of items that can be storred in collection. Defaul value is 
    :class:`KeyObject <supervisely.collection.key_indexed_collection.KeyObject>`. 
    Field has to be overridden in child class. Before adding object to collection its type is compared with 
    ``item_type`` and ``TypeError`` exception is raised if it differs. Collection is immutable.
    """
    def __init__(self, items: Optional[List[KeyObject]]=None):
        self._collection = {}
        self._add_items_impl(self._collection, take_with_default(items, []))

    def _add_impl(self, dst_collection, item):
        """
        Add given item to given collection. Raise error if type of item not KeyObject or item with an item with that name is already in given collection
        :param dst_collection: dictionary
        :param item: ObjClass, TagMeta or Tag class object
        :return: dictionary
        """
        if not isinstance(item, KeyIndexedCollection.item_type):
            raise TypeError(
                'Item type ({!r}) != {!r}'.format(type(item).__name__, KeyIndexedCollection.item_type.__name__))
        if item.key() in dst_collection:
            raise DuplicateKeyError('Key {!r} already exists'.format(item.key()))
        dst_collection[item.key()] = item

    def _add_items_impl(self, dst_collection, items):
        """
        Add items from input list to given collection. Raise error if type of item not KeyObject or item with an item with that name is already in given collection
        :param dst_collection: dictionary
        :param items: list of ObjClass, TagMeta or Tag class objects
        """
        for item in items:
            self._add_impl(dst_collection, item)

    def add(self, item: KeyObject) -> KeyIndexedCollection:
        """
        Add given item to collection.

        :param item: :class:`ObjClassCollection<supervisely.annotation.obj_class_collection.ObjClassCollection>`, :class:`TagMetaCollection<supervisely.annotation.tag_meta_collection.TagMetaCollection>`  or :class:`TagCollection<supervisely.annotation.tag_collection.TagCollection>` object.
        :type item: KeyObject
        :return: New instance of KeyIndexedCollection
        :rtype: :class:`KeyIndexedCollection<KeyIndexedCollection>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            item_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            collection = sly.collection.key_indexed_collection.KeyIndexedCollection([item_cat, item_turtle])
            # Remember that KeyIndexedCollection object is immutable, and we need to assign new instance of KeyIndexedCollection to a new variable
            item_dog = sly.ObjClass('dog', sly.Rectangle)
            new_collection = collection.add(item_dog)
        """
        return self.clone(items=[*self.items(), item])

    def add_items(self, items: List[KeyObject]) -> KeyIndexedCollection:
        """
        Add items from given list to collection.

        :param items: List of :class:`ObjClassCollection<supervisely.annotation.obj_class_collection.ObjClassCollection>`, :class:`TagMetaCollection<supervisely.annotation.tag_meta_collection.TagMetaCollection>` or :class:`TagCollection<supervisely.annotation.tag_collection.TagCollection>` objects.
        :type items:  List[KeyObject]
        :return: New instance of KeyIndexedCollection
        :rtype: :class:`KeyIndexedCollection<KeyIndexedCollection>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            item_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            collection = sly.collection.key_indexed_collection.KeyIndexedCollection([item_cat, item_turtle])
            # Remember that KeyIndexedCollection object is immutable, and we need to assign new instance of KeyIndexedCollection to a new variable
            item_dog = sly.ObjClass('dog', sly.Rectangle)
            item_mouse = sly.ObjClass('mouse', sly.Bitmap)
            new_collection = collection.add_items([item_dog, item_mouse])
        """
        return self.clone(items=[*self.items(), *items])

    def get(self, key: str, default: Optional[Any]=None) -> KeyObject:
        """
        Get item from collection with given key(name).

        :param items: Name of KeyObject in collection.
        :type items:  str
        :param default: The value that is returned if there is no key in the collection.
        :type default: optional
        :return: :class:`ObjClassCollection<supervisely.annotation.obj_class_collection.ObjClassCollection>`, :class:`TagMetaCollection<supervisely.annotation.tag_meta_collection.TagMetaCollection>` or :class:`TagCollection<supervisely.annotation.tag_collection.TagCollection>` object
        :rtype: :class:`KeyObject<KeyObject>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            item_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            collection = sly.collection.key_indexed_collection.KeyIndexedCollection([item_cat, item_turtle])

            item_cat = collection.get('cat')
            print(item_cat)
            # Output:
            # Name:  cat                      Value type:none          Possible values:None       Hotkey                  Applicable toall        Applicable classes[]

            item_not_exist = collection.get('no_item', {1: 2})
            print(item_not_exist)
            # Output:
            # {1: 2}
        """
        return self._collection.get(key, default)

    def __next__(self):
        for value in self._collection.values():
            yield value

    def __iter__(self):
        return next(self)

    def __contains__(self, item):
        return (isinstance(item, KeyIndexedCollection.item_type)
                and item == self._collection.get(item.key()))

    def __len__(self):
        return len(self._collection)

    def items(self) -> List[KeyObject]:
        """
        Get list of all items in collection.

        :return: List of :class:`ObjClassCollection<supervisely.annotation.obj_class_collection.ObjClassCollection>`, :class:`TagMetaCollection<supervisely.annotation.tag_meta_collection.TagMetaCollection>` or :class:`TagCollection<supervisely.annotation.tag_collection.TagCollection>` objects
        :rtype: :class:`List[KeyObject]`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            item_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            collection = sly.collection.key_indexed_collection.KeyIndexedCollection([item_cat, item_turtle])
            items = collection.items()
            print(items)
            # Output:
            # [<supervisely.annotation.tag_meta.TagMeta object at 0x7fd08eae4340>,
            #  <supervisely.annotation.tag_meta.TagMeta object at 0x7fd08eae4370>]
        """
        return list(self._collection.values())

    def clone(self, items: Optional[List[KeyObject]]=None) -> KeyIndexedCollection:
        """
        Makes a copy of KeyIndexedCollection with new fields, if fields are given, otherwise it will use fields of the original KeyIndexedCollection.

        :param items: List of :class:`ObjClassCollection<supervisely.annotation.obj_class_collection.ObjClassCollection>`, :class:`TagMetaCollection<supervisely.annotation.tag_meta_collection.TagMetaCollection>` or :class:`TagCollection<supervisely.annotation.tag_collection.TagCollection>` objects.
        :type items:  List[KeyObject], optional
        :return: New instance of KeyIndexedCollection
        :rtype: :class:`KeyIndexedCollection<KeyIndexedCollection>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            item_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            collection = sly.collection.key_indexed_collection.KeyIndexedCollection([item_cat, item_turtle])
            # Remember that KeyIndexedCollection object is immutable, and we need to assign new instance of KeyIndexedCollection to a new variable
            new_collection = collection.clone()
        """
        return type(self)(items=(items if items is not None else self.items()))

    def keys(self) -> List[str]:
        """
        Get list of all keys(item names) in collection.

        :return: List of collection keys
        :rtype: :class:`List[str]`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            item_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            collection = sly.collection.key_indexed_collection.KeyIndexedCollection([item_cat, item_turtle])
            keys = collection.keys() # ['cat', 'turtle']
        """
        return list(self._collection.keys())

    def has_key(self, key: str) -> bool:
        """
        Check if given key(item name exist in collection).

        :param key: The key to look for in the collection.
        :type key:  str
        :return: Is the key in the collection or not
        :rtype: :class:`bool`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            item_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            collection = sly.collection.key_indexed_collection.KeyIndexedCollection([item_cat, item_turtle])

            collection.has_key('cat') # True
            collection.has_key('hamster') # False
        """
        return key in self._collection

    def intersection(self, other: List[KeyObject]) -> KeyIndexedCollection:
        """
        Find intersection of given list of instances with collection items.

        :param key: List of :class:`ObjClassCollection<supervisely.annotation.obj_class_collection.ObjClassCollection>`, :class:`TagMetaCollection<supervisely.annotation.tag_meta_collection.TagMetaCollection>` or :class:`TagCollection<supervisely.annotation.tag_collection.TagCollection>` objects.
        :type key:  List[KeyObject]
        :raises: :class:`ValueError` if find items with same keys(item names)
        :return: KeyIndexedCollection object
        :rtype: :class:`KeyIndexedCollection<KeyIndexedCollection>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            item_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            collection = sly.collection.key_indexed_collection.KeyIndexedCollection([item_cat, item_turtle])

            item_dog = sly.TagMeta('dog', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            items = [item_dog, item_turtle]

            intersection = collection.intersection(items)
            print(intersection.to_json())
            # Output: [
            #     {
            #         "name": "turtle",
            #         "value_type": "any_string",
            #         "color": "#760F8A",
            #         "hotkey": "",
            #         "applicable_type": "all",
            #         "classes": []
            #     }
            # ]
        """
        common_items = []
        for other_item in other:
            our_item = self.get(other_item.key())
            if our_item is not None:
                if our_item != other_item:
                    raise ValueError("Different values for the same key {!r}".format(other_item.key()))
                else:
                    common_items.append(our_item)
        return self.clone(common_items)

    def difference(self, other: List[KeyObject]) -> KeyIndexedCollection:
        """
        Find difference between collection and given list of instances.

        :param key: List of :class:`ObjClassCollection<supervisely.annotation.obj_class_collection.ObjClassCollection>`, :class:`TagMetaCollection<supervisely.annotation.tag_meta_collection.TagMetaCollection>` or :class:`TagCollection<supervisely.annotation.tag_collection.TagCollection>` objects.
        :type key:  List[KeyObject]
        :return: KeyIndexedCollection object
        :rtype: :class:`KeyIndexedCollection<KeyIndexedCollection>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            item_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            collection = sly.collection.key_indexed_collection.KeyIndexedCollection([item_cat, item_turtle])

            item_dog = sly.TagMeta('dog', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            items = [item_dog, item_turtle]

            diff = collection.difference(items)
            print(diff.to_json())
            # Output: [
            #     {
            #         "name": "cat",
            #         "value_type": "none",
            #         "color": "#8A150F",
            #         "hotkey": "",
            #         "applicable_type": "all",
            #         "classes": []
            #     }
            # ]
        """
        items = [item for item in self.items() if item not in other]
        return self.clone(items)

    def merge(self, other: KeyIndexedCollection) -> KeyIndexedCollection:
        """
        Merge collection and other KeyIndexedCollection object.

        :param key: KeyIndexedCollection object.
        :type key:  KeyIndexedCollection
        :raises: :class:`ValueError` if item name from given list is in collection but items in both are different
        :return: KeyIndexedCollection object
        :rtype: :class:`KeyIndexedCollection<KeyIndexedCollection>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            item_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            collection = sly.collection.key_indexed_collection.KeyIndexedCollection([item_cat, item_turtle])

            item_dog = sly.TagMeta('dog', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            other_collection = sly.collection.key_indexed_collection.KeyIndexedCollection([item_dog, item_turtle])

            merge = collection.merge(other_collection)
            print(merge.to_json())
            # Output: [
            #     {
            #         "name": "dog",
            #         "value_type": "none",
            #         "color": "#8A6C0F",
            #         "hotkey": "",
            #         "applicable_type": "all",
            #         "classes": []
            #     },
            #     {
            #         "name": "cat",
            #         "value_type": "none",
            #         "color": "#0F4A8A",
            #         "hotkey": "",
            #         "applicable_type": "all",
            #         "classes": []
            #     },
            #     {
            #         "name": "turtle",
            #         "value_type": "any_string",
            #         "color": "#4F0F8A",
            #         "hotkey": "",
            #         "applicable_type": "all",
            #         "classes": []
            #     }
            # ]
        """
        new_items = []
        for other_item in other.items():
            our_item = self.get(other_item.key())
            if our_item is None:
                new_items.append(other_item)
            elif our_item != other_item:
                raise ValueError('Error during merge for key {!r}: values are different'.format(other_item.key()))
        return self.clone(new_items + self.items())

    def __str__(self):
        res_table = PrettyTable()
        res_table.field_names = self.item_type.get_header_ptable()
        for item in self:
            res_table.add_row(item.get_row_ptable())
        return res_table.get_string()

    def to_json(self) -> List[Dict]:
        """
        Convert the KeyIndexedCollection to a json serializable list.

        :return: List of json serializable dicts
        :rtype: :class:`List[dict]`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            item_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            collection = sly.collection.key_indexed_collection.KeyIndexedCollection([item_cat, item_turtle])
            collection_json = collection.to_json()
            # Output: [
            #     {
            #         "name": "cat",
            #         "value_type": "none",
            #         "color": "#8A0F12",
            #         "hotkey": "",
            #         "applicable_type": "all",
            #         "classes": []
            #     },
            #     {
            #         "name": "turtle",
            #         "value_type": "any_string",
            #         "color": "#8A860F",
            #         "hotkey": "",
            #         "applicable_type": "all",
            #         "classes": []
            #     }
            # ]
        """
        return [item.to_json() for item in self]

    def __eq__(self, other: KeyIndexedCollection):
        if len(self) != len(other):
            return False
        for cur_item in self:
            other_item = other.get(cur_item.key())
            if other_item is None or cur_item != other_item:
                return False
        return True

    def __ne__(self, other: KeyIndexedCollection):
        return not self == other


class MultiKeyIndexedCollection(KeyIndexedCollection):
    """
    Base class for :class:`TagCollection<supervisely.annotation.tag_collection.TagCollection>` instances. MultiKeyIndexedCollection makes it possible to add an object with an already existing key.

    :param items: List of :class:`ObjClassCollection<supervisely.annotation.obj_class_collection.ObjClassCollection>`, :class:`TagMetaCollection<supervisely.annotation.tag_meta_collection.TagMetaCollection>`  and :class:`TagCollection<supervisely.annotation.tag_collection.TagCollection>` objects.
    :type items: list, optional

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        item_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
        item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
        # Create item with same key 'cat'
        other_cat = sly.ObjClass('cat', sly.Rectangle)
        collection = sly.collection.key_indexed_collection.MultiKeyIndexedCollection([item_cat, item_turtle, other_cat])
        print(collection.to_json())
        # Output: [
        #     {
        #         "name": "cat",
        #         "value_type": "none",
        #         "color": "#0F198A",
        #         "hotkey": "",
        #         "applicable_type": "all",
        #         "classes": []
        #     },
        #     {
        #         "title": "cat",
        #         "shape": "rectangle",
        #         "color": "#0F8A6B",
        #         "geometry_config": {},
        #         "hotkey": ""
        #     },
        #     {
        #         "name": "turtle",
        #         "value_type": "any_string",
        #         "color": "#0F658A",
        #         "hotkey": "",
        #         "applicable_type": "all",
        #         "classes": []
        #     }
        # ]
    """
    def __init__(self, items: Optional[List]=None):
        self._collection = defaultdict(list)
        self._add_items_impl(self._collection, take_with_default(items, []))

    def _add_impl(self, dst_collection, item):
        if not isinstance(item, MultiKeyIndexedCollection.item_type):
            raise TypeError(
                'Item type ({!r}) != {!r}'.format(type(item).__name__, MultiKeyIndexedCollection.item_type.__name__))
        dst_collection[item.key()].append(item)

    def get(self, key: str, default: Optional[Any]=None) -> KeyObject:
        """
        Get item from collection with given key(name). If there are many values for the same key, the first value will be returned.

        :param items: Name of KeyObject in collection.
        :type items:  str
        :param default: The value that is returned if there is no key in the collection.
        :type default: optional
        :return: :class:`ObjClassCollection<supervisely.annotation.obj_class_collection.ObjClassCollection>`, :class:`TagMetaCollection<supervisely.annotation.tag_meta_collection.TagMetaCollection>` or :class:`TagCollection<supervisely.annotation.tag_collection.TagCollection>` object
        :rtype: :class:`KeyObject<KeyObject>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            item_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            other_cat = sly.ObjClass('cat', sly.Rectangle)
            collection = sly.collection.key_indexed_collection.MultiKeyIndexedCollection([item_cat, item_turtle, other_cat])
            item = collection.get('cat')
            print(item)
            # Output:
            # Name:  cat                      Value type:none          Possible values:None       Hotkey                  Applicable toall        Applicable classes[]
        """
        result = self._collection.get(key, default)
        if not result:
            return None
        return result[0]

    def get_all(self, key: str, default: Optional[List[Any]]=[]) -> List[KeyObject]:
        """
        Get item from collection with given key(name). If there are many values for the same key,all values will be returned by list.

        :param items: Name of KeyObject in collection.
        :type items:  str
        :param default: The value that is returned if there is no key in the collection.
        :type default: List, optional
        :return: List of :class:`ObjClassCollection<supervisely.annotation.obj_class_collection.ObjClassCollection>`, :class:`TagMetaCollection<supervisely.annotation.tag_meta_collection.TagMetaCollection>` or :class:`TagCollection<supervisely.annotation.tag_collection.TagCollection>` objects or empty list
        :rtype: :class:`List[KeyObject]` or :class:`list`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            item_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            other_cat = sly.ObjClass('cat', sly.Rectangle)
            collection = sly.collection.key_indexed_collection.MultiKeyIndexedCollection([item_cat, item_turtle, other_cat])
            items = collection.get('cat')
            print(items)
            # Output:
            # [<supervisely.annotation.tag_meta.TagMeta object at 0x7f0278662340>, <supervisely.annotation.obj_class.ObjClass object at 0x7f02786623a0>]
        """
        return self._collection.get(key, default)

    def __next__(self):
        for tag_list in self._collection.values():
            for tag in tag_list:
                yield tag

    def __contains__(self, item):
        return (isinstance(item, MultiKeyIndexedCollection.item_type)
                and item in self.get_all(item.key()))

    def __len__(self):
        return sum([len(tag_list) for tag_list in self._collection.values()])

    def items(self) -> List[KeyObject]:
        """
        Get list of all items in collection.

        :return: List of :class:`ObjClassCollection<supervisely.annotation.obj_class_collection.ObjClassCollection>`, :class:`TagMetaCollection<supervisely.annotation.tag_meta_collection.TagMetaCollection>` or :class:`TagCollection<supervisely.annotation.tag_collection.TagCollection>` objects
        :rtype: :class:`List[KeyObject]`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            item_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            other_cat = sly.ObjClass('cat', sly.Rectangle)
            collection = sly.collection.key_indexed_collection.MultiKeyIndexedCollection([item_cat, item_turtle, other_cat])
            print(collection.items())
            # Output:
            # [<supervisely.annotation.tag_meta.TagMeta object at 0x7fdbd28ce340>,
            #  <supervisely.annotation.obj_class.ObjClass object at 0x7fdbd28ce3a0>,
            #  <supervisely.annotation.tag_meta.TagMeta object at 0x7fdbd28ce370>]
        """
        res = []
        for tag_list in self._collection.values():
            res.extend(tag_list)
        return res

    def intersection(self, other: List[KeyObject]) -> MultiKeyIndexedCollection:
        """
        Find intersection of given list of instances with collection items.

        :param key: List of :class:`ObjClassCollection<supervisely.annotation.obj_class_collection.ObjClassCollection>`, :class:`TagMetaCollection<supervisely.annotation.tag_meta_collection.TagMetaCollection>` or :class:`TagCollection<supervisely.annotation.tag_collection.TagCollection>` objects.
        :type key:  List[KeyObject]
        :raises: :class:`ValueError` if find items with same keys(item names)
        :return: MultiKeyIndexedCollection object
        :rtype: :class:`MultiKeyIndexedCollection<MultiKeyIndexedCollection>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            item_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            other_cat = sly.ObjClass('cat', sly.Rectangle)
            collection = sly.collection.key_indexed_collection.MultiKeyIndexedCollection([item_cat, item_turtle, other_cat])

            # Note, item_cat_2 have same key as item_cat, but another value
            item_cat_2 = sly.TagMeta('cat', sly.TagValueType.ANY_STRING)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            items = [item_cat_2, item_turtle]

            intersect = collection.intersection(items)
            print(intersect.to_json())
            # Output: [
            #     {
            #         "name": "turtle",
            #         "value_type": "any_string",
            #         "color": "#5B8A0F",
            #         "hotkey": "",
            #         "applicable_type": "all",
            #         "classes": []
            #     }
            # ]
        """
        common_items = []
        for other_item in other:
            key_list = self.get_all(other_item.key())
            for our_item in key_list:
                if our_item == other_item:
                    common_items.append(our_item)
        return self.clone(common_items)

    def merge(self, other: List[KeyObject]) -> MultiKeyIndexedCollection:
        """
        Merge collection with other MultiKeyIndexedCollection object.

        :param key: MultiKeyIndexedCollection object.
        :type key:  MultiKeyIndexedCollection
        :return: MultiKeyIndexedCollection object
        :rtype: :class:`MultiKeyIndexedCollection<MultiKeyIndexedCollection>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            item_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            other_cat = sly.ObjClass('cat', sly.Rectangle)
            collection = sly.collection.key_indexed_collection.MultiKeyIndexedCollection([item_cat, item_turtle, other_cat])

            item_dog = sly.TagMeta('dog', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            other_collection = sly.collection.key_indexed_collection.MultiKeyIndexedCollection([item_dog, item_turtle])

            merge = collection.merge(other_collection)
            print(merge.to_json())
            # Output: [
            #     {
            #         "name": "cat",
            #         "value_type": "none",
            #         "color": "#198A0F",
            #         "hotkey": "",
            #         "applicable_type": "all",
            #         "classes": []
            #     },
            #     {
            #         "title": "cat",
            #         "shape": "rectangle",
            #         "color": "#898A0F",
            #         "geometry_config": {},
            #         "hotkey": ""
            #     },
            #     {
            #         "name": "turtle",
            #         "value_type": "any_string",
            #         "color": "#650F8A",
            #         "hotkey": "",
            #         "applicable_type": "all",
            #         "classes": []
            #     },
            #     {
            #         "name": "turtle",
            #         "value_type": "any_string",
            #         "color": "#0F8A83",
            #         "hotkey": "",
            #         "applicable_type": "all",
            #         "classes": []
            #     },
            #     {
            #         "name": "dog",
            #         "value_type": "none",
            #         "color": "#1A8A0F",
            #         "hotkey": "",
            #         "applicable_type": "all",
            #         "classes": []
            #     }
            # ]
        """
        new_items = [*self.items(), *other.items()]
        return self.clone(items=new_items)

    def merge_without_duplicates(self, other: MultiKeyIndexedCollection) -> MultiKeyIndexedCollection:
        """
        Merge collection with other MultiKeyIndexedCollection object. Duplicates will be ignored.

        :param key: MultiKeyIndexedCollection object.
        :type key:  MultiKeyIndexedCollection
        :raises: :class:`ValueError` if item name from given MultiKeyIndexedCollection is in collection but items in both are different
        :return: MultiKeyIndexedCollection object
        :rtype: :class:`MultiKeyIndexedCollection<MultiKeyIndexedCollection>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            item_cat = sly.TagMeta('cat', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            other_cat = sly.ObjClass('cat', sly.Rectangle)
            collection = sly.collection.key_indexed_collection.MultiKeyIndexedCollection([item_cat, item_turtle, other_cat])

            item_dog = sly.TagMeta('dog', sly.TagValueType.NONE)
            item_turtle = sly.TagMeta('turtle', sly.TagValueType.ANY_STRING)
            other_collection = sly.collection.key_indexed_collection.MultiKeyIndexedCollection([item_dog, item_turtle])

            merge = collection.merge_without_duplicates(other_collection)
            print(merge.to_json())
            # Output: [
            #     {
            #         "name": "dog",
            #         "value_type": "none",
            #         "color": "#8A0F37",
            #         "hotkey": "",
            #         "applicable_type": "all",
            #         "classes": []
            #     },
            #     {
            #         "name": "cat",
            #         "value_type": "none",
            #         "color": "#778A0F",
            #         "hotkey": "",
            #         "applicable_type": "all",
            #         "classes": []
            #     },
            #     {
            #         "title": "cat",
            #         "shape": "rectangle",
            #         "color": "#8A0F76",
            #         "geometry_config": {},
            #         "hotkey": ""
            #     },
            #     {
            #         "name": "turtle",
            #         "value_type": "any_string",
            #         "color": "#850F8A",
            #         "hotkey": "",
            #         "applicable_type": "all",
            #         "classes": []
            #     }
            # ]
        """
        return super().merge(other)
